from cvmd.registry import register_model


@register_model("deformabledetr", "deformable_detr")
class DeformableDETR:

    def __init__(self,*args,**kwargs):
        
        raise NotImplementedError("DeformableDETR is not yet implemented.")
